#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Script name: paratyping_functions.py

Contains the functions necessary for using paratyping, namely parsing parapred output and calculating paratope identity.
"""

def define_paratope(parapred_output, threshold):
    """
    Function that returns predicted paratope residues according to a chosen
    threshold. Paratyping paper uses 0.67.


    Parameters
    ----------
    parapred_output : dict
        The dictionary which is the output of the modded Parapred (first key
        is sequence ID, second key is the CDR (CDR1-3), third key is position
        within the CDR (0-len(CDR-1)), fourth key is the amino acid at that
        position (one-letter code), fifth key is Parapred's outputted prob.).
    threshold : float
        Probability threshold above which a residue is assigned "paratope" vs.
        "non-paratope".

    Returns
    -------
    paratope : dict
        A dictionary where the first key is the sequence ID, second key is the
        CDR (CDR1-3) and third key is position within the CDR (0-len(CDR)-1).
        The value of that key is the amino acid at that position (one-letter
        code).

    """
    paratope = {}
    for cdr in parapred_output:
        paratope[cdr] = {}
        for cdr_seq in parapred_output[cdr]:
            for position in parapred_output[cdr][cdr_seq]:
                for residue in parapred_output[cdr][cdr_seq][position]:
                    if parapred_output[cdr][cdr_seq][position][residue] > threshold:
                        paratope[cdr][position] = residue
    return paratope


def count_residues(paratope):
    """
    Function that counts the residues in the paratope format.

    Parameters
    ----------
    paratope : dict
        A dictionary where the first key is the sequence ID, second key is the
        CDR (CDR1-3) and third key is position within the CDR (0-len(CDR)-1).
        The value of that key is the amino acid at that position (one-letter
        code).

    Returns
    -------
    count : int
        The number of residues in the paratope.

    """   
    count = 0
    for cdr in paratope:
        for paratope_residue in paratope[cdr]:
            count += 1
    return count


def paratope_identity(paratope1, paratope2):
    """

    Parameters
    ----------
    paratope1 : dict
        First paratope in a pair to be compared.
    paratope2 : dict
        Second paratope in a pair to be compared.

    Returns
    -------
    identity : float
        Identity of paratopes 1 and 2 equal to the number of matches (residues
        at equivalent positions, predicted to be in the paratope, same residue)
        divided by the minimum number of paratope residues in the pair.

    """
    
    count = 0
    for cdr in paratope1:
        for position in paratope1[cdr]:
            if position in paratope2[cdr]:
                if paratope1[cdr][position]==paratope2[cdr][position]:
                    count += 1
    identity = float(count)/min(count_residues(paratope1),count_residues(paratope2))
    return identity
                
